using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.UI;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;


namespace RoboBrain
{
public class KalmanFilter
{
        private Emgu.CV.Kalman m_pKalmanFilter;
        private CvRandState rng;
        private CvMat* state;
        private CvMat* process_noise;
        private CvMat* measurement;
        private CvMat* measurement_noise;
        private bool m_bMeasurement;
        private clock_t m_timeLastMeasurement;
        private CriticalHandle mutexPrediction;

//   

public KalmanFilter()
{
    m_timeLastMeasurement = clock();
        int dynam_params = 4; // x,y,dx,dy
        int measure_params = 2;
        m_pKalmanFilter = cvCreateKalman(dynam_params, measure_params);
        state = cvCreateMat( dynam_params, 1, CV_32FC1 );
       
        //   
        cvRandInit( &rng, 0, 1, -1, CV_RAND_UNI );
        cvRandSetRange( &rng, 0, 1, 0 );
        rng.disttype = CV_RAND_NORMAL;
        cvRand( &rng, state );

        process_noise = cvCreateMat( dynam_params, 1, CV_32FC1 ); // (w_k)
        measurement = cvCreateMat( measure_params, 1, CV_32FC1 ); // two parameters for x,y  (z_k)
        measurement_noise = cvCreateMat( measure_params, 1, CV_32FC1 ); // two parameters for x,y (v_k)
        cvZero(measurement);

        // F matrix data
        // F is transition matrix.  It relates how the states interact
        // For single input fixed velocity the new value
        // depends on the previous value and velocity- hence 1 0 1 0
        // on top line.       
        //  ,      -  0 1 0 1 on second row
        const float[] F = {
                1, 0, 1, 0,//x + dx
                0, 1, 0, 1,//y + dy
                0, 0, 1, 0,//dx = dx
                0, 0, 0, 1,//dy = dy
        };
        memcpy( m_pKalmanFilter->transition_matrix->data.fl, F, sizeof(F));
        cvSetIdentity( m_pKalmanFilter->measurement_matrix, cvRealScalar(1) ); // (H)
        cvSetIdentity( m_pKalmanFilter->process_noise_cov, cvRealScalar(1e-5) ); // (Q)
        cvSetIdentity( m_pKalmanFilter->measurement_noise_cov, cvRealScalar(1e-1) ); // (R)
        cvSetIdentity( m_pKalmanFilter->error_cov_post, cvRealScalar(1));

        // choose random initial state
        cvRand( &rng, m_pKalmanFilter->state_post );

        
    //InitializeCriticalSection(&mutexPrediction);
}

    public ~KalmanFilter()
{
        cvReleaseMat( &state );
        cvReleaseMat( &process_noise );
        cvReleaseMat( &measurement );
        cvReleaseMat( &measurement_noise );
        cvReleaseKalman( &m_pKalmanFilter );

        //DeleteCriticalSection(&mutexPrediction);
}

    //       
    public void predictionBegin(float x, float y)
{
        //  m_pKalmanFilter->state_post   
        m_pKalmanFilter->state_post->data.fl[0] = x; // x
        m_pKalmanFilter->state_post->data.fl[1] = y; // y
        m_pKalmanFilter->state_post->data.fl[2] = (float)0; //dx -   x
        m_pKalmanFilter->state_post->data.fl[3] = (float)0; //dy -   y
}

    public void predictionUpdate(float x, float y)
{
        EnterCriticalSection(mutexPrediction);
        state->data.fl[0] = x; //center x
        state->data.fl[1] = y; //center y
        m_bMeasurement = true;
        LeaveCriticalSection(mutexPrediction);
}

    public void predictionReport(CvPoint &pnt)
{
        clock_t timeCurrent = clock();

        EnterCriticalSection(&mutexPrediction);

        const CvMat* prediction = cvKalmanPredict( m_pKalmanFilter, 0 );
        pnt.x = prediction->data.fl[0];
        pnt.y = prediction->data.fl[1];
        //      ,      
        //  .      ,    ,   
        if (m_bMeasurement) //(timeCurrent - m_timeLastMeasurement < TIME_OUT_LOCATION_UPDATE) //   
        {
                m_bMeasurement = false;

        }
        //        ..
		//  ,            
		// (    )
        else //   
        {
                state->data.fl[0] = pnt.x;
                state->data.fl[1] = pnt.y;
        }

        //    noise(z_k)
        cvRandSetRange( &rng, 0, sqrt(m_pKalmanFilter->measurement_noise_cov->data.fl[0]), 0 );
        cvRand( &rng, measurement_noise );

        // zk = Hk * xk + vk
        // measurement = measurement_error_matrix * current_state + measurement_noise
        cvMatMulAdd( m_pKalmanFilter->measurement_matrix, state, measurement_noise, measurement );

        //   
        cvKalmanCorrect( m_pKalmanFilter, measurement );

        //   (w_k)
        cvRandSetRange( &rng, 0, sqrt(m_pKalmanFilter->process_noise_cov->data.fl[0]), 0 );
        cvRand( &rng, process_noise );

        // xk = F * xk-1 + B * uk + wk
        // state = transition_matrix * previous_state + control_matrix * control_input + noise
        cvMatMulAdd( m_pKalmanFilter->transition_matrix, state, process_noise, state );
        LeaveCriticalSection(&mutexPrediction);
}
}
}