﻿using System;
using System.Collections.Generic;
using System.Text;

namespace RoboBrain
{
    public class RobotClass : Object
    {
        const int SERVO_COUNT = 15;

        private int _MaxSpeed = 60; // maximum drive speed (60 is angle in degrees like for servo)

        public ServoClass[] Servos = new ServoClass[SERVO_COUNT];

        public int MaxSpeed
        {
            get
            {
                return _MaxSpeed;
            }
            set
            {
                _MaxSpeed = value;
            }
        }

        public RobotClass()
        {
            string[] names = new string[SERVO_COUNT] {
                    "r_track", "l_track", "base",
                    "r_shoulder_rot", "r_shoulder_ud", "r_elbow", "r_wrist", "r_hand",
                    "l_shoulder_rot", "l_shoulder_ud", "l_elbow", "l_wrist", "l_hand",
                    "head_pan", "head_tilt"
                };

            for (int i = 0; i < SERVO_COUNT; i++) Servos[i] = new ServoClass( i, names[i], false, 1500, 1500, false );            
        }

        public void CheckConnection()
        {
            Transport.Cmd(scCommand.CHECK, 0x00, 0xFF00);
        }

        public void ReadPins()
        {
            for (int i = 0; i < 15; i++) Transport.Cmd(scCommand.SERVO_GET_PIN, i, 0);
        }

        public void BeginUpdate()
        {
            foreach (ServoClass s in Servos) s.BeginUpdate();
        }
        
        public void EndUpdate()
        {
            foreach (ServoClass s in Servos) s.EndUpdate();
        }

        public void BeginUpdateDrive()
        {
            Servos[0].BeginUpdate();
            Servos[1].BeginUpdate();
        }

        public void EndUpdateDrive()
        {
            Servos[0].EndUpdate();
            Servos[1].EndUpdate();
        }

        /// <summary>
        /// returns all servo positions in microseconds as string
        /// </summary>
        /// <returns>string</returns>
        public string ToStringM()
        {
            string result = "";
            foreach (ServoClass s in Servos) result += s.ToStringM() + " ";
            return result.Substring(0, result.Length - 1);
        }

        /// <summary>
        /// returns all servo positions in degrees as string
        /// </summary>
        /// <returns>string</returns>
        public string ToStringD()
        {
            string result = "";
            foreach (ServoClass s in Servos) result += s.ToStringD() + " ";
            return result.Substring(0, result.Length - 1);
        }
        
        public void EnableAllServos(bool value)
        {
            BeginUpdate();
            foreach (ServoClass s in Servos) s.Enabled = value;
            EndUpdate();
        }

        public void EnableServo(int index, bool value)
        {
            if (index >= 0 && index < SERVO_COUNT)
                Servos[index].Enabled = value;
            else
                throw new Exception("Servo " + index.ToString() + " not found");
        }

        public void SetServoSpeed(int value)
        {
            Transport.Cmd(scCommand.SERVO_GROUP_SPEED, 0, value);
        }

        public void SetServoTime(int value)
        {
            Transport.Cmd(scCommand.SERVO_GROUP_TIME, 0, value);
        }

        public void Drive(int left, int right, bool disable = false)
        {
            if (disable)
            {
                Servos[1].Enabled = false;
                Servos[0].Enabled = false;
            }

            if (left > _MaxSpeed) left = _MaxSpeed;
            if (right > _MaxSpeed) right = _MaxSpeed;
            Servos[0].Position = right;
            Servos[1].Position = left;
            
            if (Servos[0].GetAngle() != 0 || Servos[1].GetAngle() != 0)
            {
                Servos[1].Enabled = true;
                Servos[0].Enabled = true;
            }
        }

        public void OffsetServoM(string name, int value)
        {
            int index = GetServoIndexByName(name);
            if (index >= 0 && index < SERVO_COUNT)
            {
                Servos[index].OffsetM(value);
                Servos[index].Enabled = true;
            }
            else
                throw new Exception("Servo \"" + name + "\" not found");
        }

        public void OffsetServoD(string name, int value)
        {
            int index = GetServoIndexByName(name);
            if (index >= 0 && index < SERVO_COUNT)
            {
                Servos[index].OffsetD(value);
                Servos[index].Enabled = true;
            }
            else
                throw new Exception("Servo \"" + name + "\" not found");
        }

        public int GetServoAngle(string name)
        {
            int index = GetServoIndexByName(name);
            if (index >= 0 && index < SERVO_COUNT) return Servos[index].GetAngle();
            else throw new Exception("Servo \"" + name + "\" not found");
        }

        public bool GetServoEnabled(string name)
        {
            int index = GetServoIndexByName(name);
            if (index >= 0 && index < SERVO_COUNT) return Servos[index].Enabled;
            else throw new Exception("Servo \"" + name + "\" not found");
        }

        public void PositionServo(string name, int position)
        {
            int index = GetServoIndexByName(name);
            if (index >= 0 && index < SERVO_COUNT)
            {
                Servos[index].Position = position;
                Servos[index].Enabled = true;
            }
            else
                throw new Exception("Servo \"" + name + "\" not found");
        }
        
        public void PositionServo(int index, int position)
        {
            if (index >= 0 && index < SERVO_COUNT)
            {
                Servos[index].Position = position;
                Servos[index].Enabled = true;
            }
            else
                throw new Exception("Invalid servo index " + index.ToString());
        }
        
        public int GetServoIndexByName(string name)
        {
            for (int i = 0; i < SERVO_COUNT; i++)
                if (Servos[i].Name.Equals(name)) return i;
            return -1;
        }

        public void DigitalWrite(int pin, int value)
        {
            Transport.Cmd(scCommand.PIN_MODE, pin, 0);
            Transport.Cmd(scCommand.DIGITAL_WRITE, pin, value);
        }

        public void SaveToIni()
        {
            foreach (ServoClass s in Servos) s.SaveToIni();
        }
        
        public void ReadFromIni()
        {
            foreach (ServoClass s in Servos) s.ReadFromIni();
        }
    }
}
