using System;
using System.IO.Ports;
using System.Threading;
using System.Text;

namespace RoboBrain
{
    public enum scCommand : byte
    {
        DIGITAL_WRITE = 0x01,
        DIGITAL_READ = 0x03,
        ANALOG_READ = 0x04,
        PIN_MODE = 0x05,
        
        SERVO_SET_POSITION = 0x81,
        SERVO_GET_POSITION = 0x82,
        SERVO_SET_SPEED = 0x83,
        SERVO_GET_SPEED = 0x84,
        SERVO_SET_TIME = 0x85,
        SERVO_GET_TIME = 0x86,
        SERVO_STOP = 0x87,
        SERVO_DISABLE = 0x88,
        SERVO_MOVING = 0x89,
        SERVO_SET_PIN = 0x8C,
        SERVO_GET_PIN = 0x8D,
        SERVO_SET_ZERO = 0x8E,
        SERVO_GET_ZERO = 0x8F,
        SERVO_INIT = 0x90,
        SERVO_DETACH = 0x91,
        RESET_GAMEPAD = 0x92,
        
        SERVO_GROUP_SPEED = 0xA0,
        SERVO_GROUP_TIME = 0xA1,
        SERVO_GROUP_MOVING = 0xA2,

        UPTIME = 0xFE,
        CHECK = 0xFF
    }

    public static class Transport
    {
        public static bool active = false;
        public static bool connected = false;
        public static uint uptime = 0;
        public static double[] analog = new double[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        public static int[] servo_position = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        public static int[] servo_pin = new int[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
        public static bool logData = false;
        private static Thread readThread = null;
        public static SerialPort com = null;
        public delegate void OnMessageHandler(object o, string message);
        public static event OnMessageHandler OnMessage; // non thread-safe
        private static int CheckParam1 = 0;
        private static int CheckParam2 = 0xFF00;
                
        public static void Open(string PortName)
        {
            readThread = new Thread(new ThreadStart(Read));
            if (com == null) com = new SerialPort();

            com.Close();
            com.PortName = PortName;
            com.BaudRate = 115200;
            com.DataBits = 8;
            com.Parity = Parity.None;
            com.StopBits = StopBits.One;
            com.Handshake = Handshake.None;
            com.ReadTimeout = 500;
            com.WriteTimeout = 500;
            com.RtsEnable = false;
            com.DtrEnable = false;
            com.Encoding = ASCIIEncoding.ASCII;
            com.DiscardNull = false;
            
            connected = false;
            com.Open();
            active = true;
            
            readThread.Start();            
        }

        public static void Close()
        {
            if (active)
            {
                active = false;
                connected = false;
                uptime = 0;
                for (int i = 0; i < analog.Length; i++) analog[i] = 0;
                for (int i = 0; i < servo_position.Length; i++) servo_position[i] = 0;
                readThread.Join();
                com.Close();
            }
        }

        public static void Cmd(scCommand cmd, int param1, int param2)
        {
            if (active)
            {
                if (cmd == scCommand.CHECK)
                {
                    CheckParam1 = param1;
                    CheckParam2 = param2;
                }
                byte[] buf = new byte[4];
                buf[0] = (byte)cmd;
                buf[1] = (byte)param1;
                buf[2] = (byte)param2;
                buf[3] = (byte)(param2 >> 8);
                Send(buf);
            }
        }

        private static void Send(byte[] buf)
        {
            try
            {
                if (active && buf != null) com.Write(buf, 0, buf.Length);
            }
            catch (Exception ex)
            {
                if (OnMessage != null) OnMessage(com, "send() -> " + ex.Message);
                Close();
            }
        }

        private static void Read()
        {
            while (active)
            {
                try
                {
                    if (com.IsOpen && com.BytesToRead > 3)
                    {
                        byte[] buf = new byte[4];
                        com.Read(buf, 0, 4);
                        OnData(buf);
                    }
                }
                catch (Exception) { }
            }
        }

        private static void OnData(byte[] buf)
        {
            try
            {
                scCommand cmd = (scCommand)buf[0];
                int param1 = (int)buf[1];
                int param2 = (int)(buf[2] + buf[3] * 256);

                if (cmd == scCommand.CHECK) connected = param1 == CheckParam1 && param2 == CheckParam2;
                else if (cmd == scCommand.SERVO_GET_POSITION) servo_position[param1] = param2;
                else if (cmd == scCommand.SERVO_GET_PIN) servo_pin[param1] = param2;
                else if (cmd == scCommand.ANALOG_READ) analog[param1] = param2 * 0.0049;
                else if (cmd == scCommand.UPTIME) if (param1 == 0) uptime = (uint)param2; else uptime += (uint)param2 << 16;

                if (logData)
                {
                    string msg = "";
                    if (cmd == scCommand.SERVO_GET_POSITION) msg = "Servo " + param1.ToString() + " position " + param2.ToString();
                    else if (cmd == scCommand.SERVO_GET_SPEED) msg = "Servo " + param1.ToString() + " speed " + param2.ToString();
                    else if (cmd == scCommand.SERVO_GET_TIME) msg = "Servo " + param1.ToString() + " time " + param2.ToString();
                    else if (cmd == scCommand.SERVO_GET_PIN) msg = "Servo " + param1.ToString() + " pin " + param2.ToString();
                    else if (cmd == scCommand.SERVO_MOVING)
                    {
                        if (param2 == 0) msg = "Servo " + param1.ToString() + " not moving";
                        else msg = "Servo " + param1.ToString() + " moving";
                    }
                    else if (cmd == scCommand.SERVO_GROUP_MOVING)
                    {
                        if (param1 == 0xFF) msg = "No servos moving";
                        else msg = "Servo " + param1.ToString() + " moving";
                    }
                    else if (cmd == scCommand.DIGITAL_READ) msg = "Digital pin " + param1.ToString() + " = " + param2.ToString();
                    else if (cmd == scCommand.ANALOG_READ)
                    {
                        msg = "Analog pin " + param1.ToString() + " = " + (param2 * 0.0049).ToString() + "V";
                    }
                    else if (cmd == scCommand.CHECK)
                    {
                        if (param1 == CheckParam1 && param2 == CheckParam2) msg = "Connection OK";
                        else msg = "Connection ERR";
                    }
                    else msg = "cmd = " + ((int)cmd).ToString() + ";index = " + param1.ToString() + "; param = " + param2.ToString();

                    msg = "[" + ByteArrayToHex(buf) + "] " + msg;

                    if (msg.Length > 0 && OnMessage != null) OnMessage(com, msg);
                }
            }
            catch (Exception ex)
            {
                if (OnMessage != null) OnMessage(com, "Error in OnData. " + ex.Message);
            }
        }

        private static string ByteArrayToHex(byte[] bytes)
        {
            StringBuilder result = new StringBuilder();
            string hex = "0123456789ABCDEF";
            foreach (byte b in bytes)
            {
                result.Append(hex[(int)b >> 4]);
                result.Append(hex[(int)b & 0xF]);
            }
            return result.ToString();
        }

    }
}