﻿using System;
using System.Collections.Generic;
using System.Text;

namespace RoboBrain
{
    public class ServoClass : Object
    {
        const int MIN_SERVO_POS         = 500;
        const int MAX_SERVO_POS         = 2500;
        
        private int _scindex;
        private bool _enabled;
        private int _position;
        private int _nullPosition;
        private int _minPosition;
        private int _maxPosition;
        private int _mcsecIn90Degrees;
        private bool _inverse;
        private string _name;
        private bool _updating;
        private bool _changed;
        
        public ServoClass(int scindex, string name, bool enabled, int position, int null_position, bool inverse)
        {
            _scindex = scindex;
            _enabled = enabled;
            _position = position;
            _nullPosition = null_position;
            _minPosition = MIN_SERVO_POS;
            _maxPosition = MAX_SERVO_POS;
            _mcsecIn90Degrees = 900;
            _inverse = inverse;
            _name = name;
            _updating = false;
            _changed = false;
        }

        ~ServoClass()
        {
            _enabled = false;
        }

        public string Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }

        /// <summary>
        /// index in servocontroller servo array
        /// </summary>
        public int SCIndex
        {
            get
            {
                return _scindex;
            }
            set
            {
                _scindex = value;
            }
        }

        public bool Enabled
        {
            get
            {
                return _enabled;
            }
            set
            {
                if (value != _enabled)
                {
                    _enabled = value;
                    _changed = true;
                    if (!_updating) SendCommand();
                }
            }
        }

        public bool Inverse
        {
            get
            {
                return _inverse;
            }
            set
            {
                if (value != _inverse)
                {
                    _inverse = value;
                    _changed = true;
                    if (!_updating) SendCommand();
                }
            }
        }

        public int Position
        {
            get
            {
                return _position;
            }
            set
            {
                if (value < 360)
                {
                    if (_inverse) value = -value;
                    value = _nullPosition + value * _mcsecIn90Degrees / 90;
                }
                if (value < _minPosition) value = _minPosition;
                if (value > _maxPosition) value = _maxPosition;
                if (value != _position)
                {
                    _position = value;
                    _changed = true;
                    if (!_updating) SendCommand();
                }
            }
        }

        public int MinPosition
        {
            get
            {
                return _minPosition;
            }
            set
            {
                if (value != _minPosition)
                {
                    _minPosition = value;
                    Position = _position; // for range check
                }
            }
        }

        public int MaxPosition
        {
            get
            {
                return _maxPosition;
            }
            set
            {
                if (value != _maxPosition)
                {
                    _maxPosition = value;
                    Position = _position; // for range check
                }
            }
        }

        public int NullPosition
        {
            get
            {
                return _nullPosition;
            }
            set
            {
                _nullPosition = value;
            }
        }

        /// <summary>
        /// number of microseconds in 90 degrees (900 for hitec hs-645mg)
        /// </summary>
        public int McsecIn90Degrees
        {
            get
            {
                return _mcsecIn90Degrees;
            }
            set
            {
                if (value != _mcsecIn90Degrees)
                {
                    _mcsecIn90Degrees = value;
                    Position = _position; // for range check
                }
            }
        }

        public int GetAngle()
        {
            if (_inverse) return -(_position - _nullPosition) * 90 / _mcsecIn90Degrees;
            else return (_position - _nullPosition) * 90 / _mcsecIn90Degrees;
        }

        public void OffsetM(int value)
        {
            if (_inverse) value = -value;
            Position = Position + value;
        }

        public void OffsetD(int value)
        {
            if (_inverse) value = -value;
            Position = _position + value * _mcsecIn90Degrees / 90;
        }

        public string ToStringM()
        {
            if (_enabled) return _scindex.ToString() + "=" + _position.ToString();
            else return _scindex.ToString();
        }

        public string ToStringD()
        {
            if (_enabled) return _scindex.ToString() + "=" + GetAngle().ToString();
            else return _scindex.ToString();
        }

        public void BeginUpdate()
        {
            _changed = false;
            _updating = true;
        }

        public void EndUpdate(bool force = false)
        {
            SendCommand(force);
        }

        private void SendCommand(bool force = false)
        {
            if (_changed || force)
            {
                if (_enabled) Transport.Cmd(scCommand.SERVO_SET_POSITION, _scindex, _position);
                else Transport.Cmd(scCommand.SERVO_DISABLE, _scindex, 0);
            }
            _changed = false;
            _updating = false;
        }

        public void SaveToIni()
        {
            IniClass.Section = _name;
            IniClass.Write("SCIndex", _scindex);
            IniClass.Write("Position", _position);
            IniClass.Write("NullPosition", _nullPosition);
            IniClass.Write("MinPosition", _minPosition);
            IniClass.Write("MaxPosition", _maxPosition);
            IniClass.Write("MsecIn90Degrees", _mcsecIn90Degrees);
            IniClass.Write("Inverse", _inverse);
        }

        public void ReadFromIni()
        {
            IniClass.Section = _name;
            _inverse =          IniClass.Read("Inverse", Inverse);
            _scindex =          IniClass.Read("SCIndex", SCIndex);
            _nullPosition =     IniClass.Read("NullPosition", NullPosition);
            _minPosition =      IniClass.Read("MinPosition", MinPosition);
            _maxPosition =      IniClass.Read("MaxPosition", MaxPosition);
            _mcsecIn90Degrees = IniClass.Read("MsecIn90Degrees", McsecIn90Degrees);
            Position =          IniClass.Read("Position", Position);
        }

    }
}
